import os
import sys
import time
import numpy as np
import re

if __name__ == "__main__":
   sys.path.append(os.path.join('..', '..', '..'))


import_pending = True


def import_all():
    global MMV15, import_pending
    global MM0,  MMV, MMVector, MMSpace
    global OrderVectorMod15
    from mmgroup.structures.mm0_group import MM0
    from mmgroup.mm_space import MMV, MMVector, MMSpace
    from mmgroup.dev.mm_reduce.order_vector import OrderVectorMod15
    MMV15 = MMV(15)
    import_pending = False
  


def compress_order_vector_data(ov):
    t = []
    v = ov.data
    def compress24(start, length):
        for i in range(start, start + 2*length, 2):
             t.append(int(v[i]) & 0xffffffff)
             t.append(int(v[i]) >> 32)
             t.append(int(v[i+1]) & 0xffffffff)
    def compress64(start, length):
        for i in range(start, start + 4*length):
             t.append(int(v[i]) & 0xffffffff)
             t.append(int(v[i]) >> 32)
    OFS_ABC = 0
    OFS_T = 72*2
    OFS_XYZ = OFS_T + 759*4
    compress24(OFS_ABC, 72)
    compress64(OFS_T, 759)
    compress24(OFS_XYZ, 3*2048)
    return np.array(t, dtype = np.uint32)


#####################################################################
# Relations in Frobenius group of order 21 generated by tags 't', 'l'
#####################################################################

def tlt_element(n, tags):
    a = [(t, 1 + ((n >> j) & 1)) for j, t in enumerate(tags)]
    return MM0(a)


def relations_tlt(tags, v):
    d = {}
    for n in range(1 << len(tags)):
        d[n] = (v * tlt_element(n, tags)).hash()
    return d

def tlt_to_ltl(verbose =  0):
    if verbose:
        s = "Make table with map  t^i l^j t^k -> l^i' t^j' l^k'" 
        print(s)
    if import_pending:
        import_all()
    for _ in range(100):
        v = MMV15('R')
        src = relations_tlt("tlt", v)
        dest = relations_tlt("ltl", v)
        h_src, h_dest = src.values(), dest.values()
        if len(h_src) == 8 and set(h_src) == set(h_dest):
            rev_dest = dict([(h, k) for k, h in dest.items()])
            d_out = dict([(k, rev_dest[src[k]]) for k in src]) 
            v1 = MMV15('R')
            for j in range(8):
                tlt = tlt_element(j, "tlt")
                ltl = tlt_element(d_out[j], "ltl")
                assert v1* tlt == v1 *ltl
            if verbose:
                print(d_out)
            return d_out
        if verbose:
            print("Attempt failed")
    raise ValueError("tlt->ltl tag conversion has failed")

def tlt_conversion_table():
    d = tlt_to_ltl()
    return sum(d[i] << (4*i) for i in range(8))

#####################################################################
# End of dealing with in Frobenius group of order 21
#####################################################################



class OrderVectorTable:
    directives = {}
    tables = None

    def compute_data(self):
        """Compute data for table classes

        We cannot to this ab initio; because the required inputs 
        are not available if sys.arg contains the argument ``mockup``.
        """
        if not self.__class__.tables is None:
            return
        if import_pending:
            import_all()
        ov_obj = OrderVectorMod15('py')
        ov = ov_obj.order_vector
        ov_hash = ov.hash()  
        a_ov =  compress_order_vector_data(ov)
        tag_data = ov_obj.tag_data
        tlt_conversion = tlt_conversion_table()
        self.__class__.tables = {
            "ORDER_VECTOR_DATA": a_ov,
            "ORDER_VECTOR_TAG_ENUM_OFFSETS": ov_obj.enum_comments(),
            "ORDER_VECTOR_TAG_DATA": tag_data,
            "ORDER_VECTOR_HASH": ov_hash,
            "ORDER_VECTOR_TLT_CONVERSION": tlt_conversion,
            }
    
    def __init__(self, *args, **kwds):
        self.compute_data()


class Mockup_OrderVectorTable:
    def __init__(self, *args, **kwds):
        pass
    a_ov =  np.array([0], dtype = np.uint32)
    tag_data = np.array([0], dtype = np.uint32)
    tables = {
        "ORDER_VECTOR_DATA": a_ov,
        "ORDER_VECTOR_TAG_ENUM_OFFSETS": "OFS_DUMMY : 0",
        "ORDER_VECTOR_TAG_DATA": tag_data,
        "ORDER_VECTOR_HASH": 0,
        "ORDER_VECTOR_TLT_CONVERSION": 0,
    }
    directives = {}


    


Tables = OrderVectorTable
MockupTables = Mockup_OrderVectorTable



              

if __name__ == "__main__":
   tlt_to_ltl(verbose = 1)
   print ("Hex table:", hex(tlt_conversion_table()))

